part of '../flutterx_live_data.dart';

/// [LiveData] which represent a collection
///
/// Every subclass of [LiveCollection] must implement [C] and comes a proxy between the observers and it's source
/// Every single change to the collection will [invalidate] this LiveData and notify it's observers
///
/// The source is passed to the constructor and then is not accessible directly from this class (there is no
/// source accessor). Anyway the value received by the [Observer] is the source instance and not this LiveData.
/// This means that if you want to update the collection from the observer you have to reference this instance
/// instead of using [Observer] received value
///
/// To update the entire collection you can call [setTo] which has the function to clear the source and copy
/// the new value into it, anyway the source itself can never be reassigned
abstract class LiveCollection<C> extends LiveData<C> {
  /// Create a LiveCollection from [source]
  LiveCollection(C source) {
    super.setData(source);
  }

  @override
  @Deprecated('Use postSetTo')
  void postData(FutureOr<C> value) => super.postData(value);

  @override
  @Deprecated('Use setTo')
  void setData(C value) => setTo(value);

  /// Replace entire collection content with [value]
  void setTo(C value);

  /// Calls [setTo] after future task completion
  void postSetTo(FutureOr<C> value) => super.postData(value);
}

/// A [LiveCollection] which implements [List]
class LiveList<E> extends LiveCollection<List<E>> implements List<E> {
  /// Create from [source]
  LiveList({List<E>? source}) : super(source ?? <E>[]);

  @override
  bool shouldUpdateValue(List<E> oldValue, List<E> newValue) => !listEquals(oldValue, newValue);

  @override
  void setTo(Iterable<E> value) {
    _data
      ..clear()
      ..addAll(value);
    invalidate();
  }

  @override
  bool any(bool Function(E) test) => _data.any(test);

  @override
  List<T> cast<T>() => _data.cast<T>();

  @override
  bool contains(Object? element) => _data.contains(element);

  @override
  E elementAt(int index) => _data.elementAt(index);

  @override
  bool every(bool Function(E) test) => _data.every(test);

  @override
  Iterable<T> expand<T>(Iterable<T> Function(E) f) => _data.expand(f);

  @override
  E get first => _data.first;

  @override
  E firstWhere(bool Function(E) test, {E Function()? orElse}) => _data.firstWhere(test, orElse: orElse);

  @override
  T fold<T>(T initialValue, T Function(T previousValue, E element) combine) => _data.fold(initialValue, combine);

  @override
  Iterable<E> followedBy(Iterable<E> other) => _data.followedBy(other);

  @override
  void forEach(void Function(E) f) => _data.forEach(f);

  @override
  bool get isEmpty => _data.isEmpty;

  @override
  bool get isNotEmpty => _data.isNotEmpty;

  @override
  Iterator<E> get iterator => _data.iterator;

  @override
  String join([String separator = '']) => _data.join(separator);

  @override
  E get last => _data.last;

  @override
  E lastWhere(bool Function(E) test, {E Function()? orElse}) => _data.lastWhere(test, orElse: orElse);

  @override
  int get length => _data.length;

  @override
  Iterable<T> map<T>(T Function(E) f) => _data.map(f);

  @override
  E reduce(E Function(E value, E element) combine) => _data.reduce(combine);

  @override
  E get single => _data.single;

  @override
  E singleWhere(bool Function(E) test, {E Function()? orElse}) => _data.singleWhere(test, orElse: orElse);

  @override
  Iterable<E> skip(int n) => _data.skip(n);

  @override
  Iterable<E> skipWhile(bool Function(E) test) => _data.skipWhile(test);

  @override
  Iterable<E> take(int n) => _data.take(n);

  @override
  Iterable<E> takeWhile(bool Function(E) test) => _data.takeWhile(test);

  @override
  List<E> toList({bool growable = true}) => _data.toList(growable: growable);

  @override
  Set<E> toSet() => _data.toSet();

  @override
  Iterable<E> where(bool Function(E) test) => _data.where(test);

  @override
  Iterable<T> whereType<T>() => _data.whereType<T>();

  @override
  String toString() => _data.toString();

  @override
  E operator [](int index) => _data[index];

  @override
  void operator []=(int index, E value) {
    if (identical(_data[index], value)) return;
    _data[index] = value;
    invalidate();
  }

  @override
  List<E> operator +(List<E> other) => _data + other;

  @override
  void add(E value) {
    _data.add(value);
    invalidate();
  }

  @override
  void addAll(Iterable<E> iterable) {
    if (iterable.isEmpty) return;
    _data.addAll(iterable);
    invalidate();
  }

  @override
  Map<int, E> asMap() => _data.asMap();

  @override
  void clear() {
    if (_data.isEmpty) return;
    _data.clear();
    invalidate();
  }

  @override
  void fillRange(int start, int end, [E? fillValue]) {
    if (start == end) return;
    _data.fillRange(start, end, fillValue);
    invalidate();
  }

  @override
  set first(E value) {
    if (identical(_data.first, value)) return;
    _data.first = value;
    invalidate();
  }

  @override
  Iterable<E> getRange(int start, int end) => _data.getRange(start, end);

  @override
  int indexOf(E element, [int start = 0]) => _data.indexOf(element, start);

  @override
  int indexWhere(bool Function(E) test, [int start = 0]) => _data.indexWhere(test, start);

  @override
  void insert(int index, E element) {
    _data.insert(index, element);
    invalidate();
  }

  @override
  void insertAll(int index, Iterable<E> iterable) {
    if (iterable.isEmpty) return;
    _data.insertAll(index, iterable);
    invalidate();
  }

  @override
  set last(E value) {
    if (identical(_data.last, value)) return;
    _data.last = value;
    invalidate();
  }

  @override
  int lastIndexOf(E element, [int? start]) => _data.lastIndexOf(element, start);

  @override
  int lastIndexWhere(bool Function(E) test, [int? start]) => _data.lastIndexWhere(test, start);

  @override
  set length(int newLength) {
    if (_data.length == newLength) return;
    _data.length = newLength;
    invalidate();
  }

  @override
  bool remove(Object? value) {
    final result = _data.remove(value);
    if (result) invalidate();
    return result;
  }

  @override
  E removeAt(int index) {
    final result = _data.removeAt(index);
    invalidate();
    return result;
  }

  @override
  E removeLast() {
    final result = _data.removeLast();
    invalidate();
    return result;
  }

  @override
  void removeRange(int start, int end) {
    if (start == end) return;
    _data.removeRange(start, end);
    invalidate();
  }

  @override
  void removeWhere(bool Function(E) test) {
    final length = _data.length;
    _data.removeWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  void replaceRange(int start, int end, Iterable<E> iterable) {
    if (start == end && iterable.isEmpty) return;
    _data.replaceRange(start, end, iterable);
    invalidate();
  }

  @override
  void retainWhere(bool Function(E) test) {
    final length = _data.length;
    _data.retainWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  Iterable<E> get reversed => _data.reversed;

  @override
  void setAll(int index, Iterable<E> iterable) {
    if (index == _data.length || iterable.isEmpty) return;
    _data.setAll(index, iterable);
    invalidate();
  }

  @override
  void setRange(int start, int end, Iterable<E> iterable, [int skipCount = 0]) {
    if (start == end) return;
    _data.setRange(start, end, iterable, skipCount);
    invalidate();
  }

  @override
  void shuffle([Random? random]) {
    _data.shuffle(random);
    invalidate();
  }

  @override
  void sort([int Function(E, E)? compare]) {
    _data.sort(compare);
    invalidate();
  }

  @override
  List<E> sublist(int start, [int? end]) => _data.sublist(start, end);
}

/// [MediatorMixin] implementation of [LiveList]
class MediatorLiveList<E> extends LiveList<E> with MediatorMixin<List<E>> {
  @override
  final bool reactive;

  /// Check [LiveList.new]
  MediatorLiveList({List<E>? source, this.reactive = true}) : super(source: source);
}

/// [HistoryMixin] implementation of [LiveList]
class HistoryLiveList<E> extends LiveList<E> with HistoryMixin<List<E>> {
  /// Check [LiveList.new]
  HistoryLiveList({List<E>? source}) : super(source: source);
}

/// [StorageMixin] implementation of [LiveList]
class StorageLiveList<E> extends LiveList<E> with StorageMixin<List<E>> {
  @override
  final String key;

  @override
  final List<E> defaultValue;

  @override
  final Map<String, String> storage;

  @override
  final String Function(List<E> value) encode;

  /// Check [StorageMixin]
  StorageLiveList({
    required this.key,
    required this.storage,
    List<E>? defaultValue,
    this.encode = jsonEncode,
    List<E> Function(String source)? decode,
  })  : defaultValue = defaultValue ?? const [],
        super(
            source:
                StorageMixin.resolve(key, storage, defaultValue?.toList(growable: true), decode ?? _jsonDecodeList));

  /// Check [StorageMixin]
  StorageLiveList.enumeration({
    required String key,
    required Map<String, String> storage,
    List<E>? defaultValue,
    required Iterable<E> values,
  }) : this(
            key: key,
            storage: storage,
            defaultValue: defaultValue,
            encode: StorageMixin.encodeEnumIterable,
            decode: (source) => StorageMixin.decodeEnumIterable<E>(source, values).toList(growable: true));

  /// Check [StorageMixin]
  StorageLiveList.jsonObject({
    required String key,
    required Map<String, String> storage,
    List<E>? defaultValue,
    dynamic Function(E value)? toJson,
    required E Function(Map<String, dynamic> json) fromJson,
  }) : this(
            key: key,
            storage: storage,
            defaultValue: defaultValue,
            encode: (value) => StorageMixin.encodeJsonIterable(toJson == null ? value : value.map(toJson)),
            decode: (source) =>
                StorageMixin.decodeJsonIterable<E>(source, (item) => fromJson(item)).toList(growable: true));

  static List<E> _jsonDecodeList<E>(String source) => List<E>.from(jsonDecode(source));
}

/// A [LiveCollection] which implements [Map]
class LiveMap<K, V> extends LiveCollection<Map<K, V>> implements Map<K, V> {
  /// Create from [source]
  LiveMap({Map<K, V>? source}) : super(source ?? <K, V>{});

  @override
  bool shouldUpdateValue(Map<K, V> oldValue, Map<K, V> newValue) => !mapEquals(oldValue, newValue);

  @override
  void setTo(Map<K, V> value) {
    _data
      ..clear()
      ..addAll(value);
    invalidate();
  }

  @override
  V? operator [](Object? key) => _data[key];

  @override
  void operator []=(K key, V value) {
    if (identical(_data[key], value)) return;
    _data[key] = value;
    invalidate();
  }

  @override
  void addAll(Map<K, V> other) {
    if (other.isEmpty) return;
    _data.addAll(other);
    invalidate();
  }

  @override
  void addEntries(Iterable<MapEntry<K, V>> entries) {
    if (entries.isEmpty) return;
    _data.addEntries(entries);
    invalidate();
  }

  @override
  void clear() {
    if (_data.isEmpty) return;
    _data.clear();
    invalidate();
  }

  @override
  Map<K2, V2> cast<K2, V2>() => _data.cast<K2, V2>();

  @override
  bool containsKey(Object? key) => _data.containsKey(key);

  @override
  bool containsValue(Object? value) => _data.containsValue(value);

  @override
  Iterable<MapEntry<K, V>> get entries => _data.entries;

  @override
  void forEach(void Function(K, V) f) => _data.forEach(f);

  @override
  bool get isEmpty => _data.isEmpty;

  @override
  bool get isNotEmpty => _data.isNotEmpty;

  @override
  Iterable<K> get keys => _data.keys;

  @override
  int get length => _data.length;

  @override
  Map<K2, V2> map<K2, V2>(MapEntry<K2, V2> Function(K, V) transform) => _data.map(transform);

  @override
  V putIfAbsent(K key, V Function() ifAbsent) {
    var updated = false;
    final result = _data.putIfAbsent(key, () {
      updated = true;
      return ifAbsent();
    });
    if (updated) invalidate();
    return result;
  }

  @override
  V? remove(Object? key) {
    final length = _data.length;
    final result = _data.remove(key);
    if (length != _data.length) invalidate();
    return result;
  }

  @override
  void removeWhere(bool Function(K, V) test) {
    final length = _data.length;
    _data.removeWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  Iterable<V> get values => _data.values;

  @override
  String toString() => _data.toString();

  @override
  V update(K key, V Function(V) update, {V Function()? ifAbsent}) {
    var updated = false;
    final result = _data.update(key, (value) {
      final newValue = update(value);
      updated = !identical(value, newValue);
      return newValue;
    },
        ifAbsent: ifAbsent == null
            ? null
            : () {
                updated = true;
                return ifAbsent();
              });
    if (updated) invalidate();
    return result;
  }

  @override
  void updateAll(V Function(K, V) update) {
    var updated = false;
    _data.updateAll((key, value) {
      final newValue = update(key, value);
      updated |= !identical(value, newValue);
      return newValue;
    });
    if (updated) invalidate();
  }
}

/// [MediatorMixin] implementation of [LiveMap]
class MediatorLiveMap<K, V> extends LiveMap<K, V> with MediatorMixin<Map<K, V>> {
  @override
  final bool reactive;

  /// Check [LiveMap.new]
  MediatorLiveMap({Map<K, V>? source, this.reactive = true}) : super(source: source);
}

/// [MediatorMixin] implementation of [LiveMap]
class HistoryLiveMap<K, V> extends LiveMap<K, V> with HistoryMixin<Map<K, V>> {
  /// Check [LiveMap.new]
  HistoryLiveMap({Map<K, V>? source}) : super(source: source);
}

/// [StorageMixin] implementation of [LiveMap]
class StorageLiveMap<K, V> extends LiveMap<K, V> with StorageMixin<Map<K, V>> {
  @override
  final String key;

  @override
  final Map<K, V> defaultValue;

  @override
  final Map<String, String> storage;

  @override
  final String Function(Map<K, V> value) encode;

  /// Check [StorageMixin]
  StorageLiveMap({
    required this.key,
    required this.storage,
    Map<K, V>? defaultValue,
    this.encode = jsonEncode,
    Map<K, V> Function(String source)? decode,
  })  : defaultValue = defaultValue ?? const {},
        super(
            source: StorageMixin.resolve(
                key, storage, defaultValue == null ? null : Map.of(defaultValue), decode ?? _jsonDecodeMap));

  static Map<K, V> _jsonDecodeMap<K, V>(String source) => Map<K, V>.from(jsonDecode(source));
}

/// A [LiveCollection] which implements [Set]
class LiveSet<E> extends LiveCollection<Set<E>> implements Set<E> {
  /// Create from [source]
  LiveSet({Set<E>? source}) : super(source ?? <E>{});

  @override
  bool shouldUpdateValue(Set<E> oldValue, Set<E> newValue) => !setEquals(oldValue, newValue);

  @override
  void setTo(Iterable<E> value) {
    _data
      ..clear()
      ..addAll(value);
    invalidate();
  }

  @override
  bool any(bool Function(E) test) => _data.any(test);

  @override
  Set<T> cast<T>() => _data.cast<T>();

  @override
  bool contains(Object? element) => _data.contains(element);

  @override
  E elementAt(int index) => _data.elementAt(index);

  @override
  bool every(bool Function(E) test) => _data.every(test);

  @override
  Iterable<T> expand<T>(Iterable<T> Function(E) f) => _data.expand(f);

  @override
  E get first => _data.first;

  @override
  E firstWhere(bool Function(E) test, {E Function()? orElse}) => _data.firstWhere(test, orElse: orElse);

  @override
  T fold<T>(T initialValue, T Function(T previousValue, E element) combine) => _data.fold(initialValue, combine);

  @override
  Iterable<E> followedBy(Iterable<E> other) => _data.followedBy(other);

  @override
  void forEach(void Function(E) f) => _data.forEach(f);

  @override
  bool get isEmpty => _data.isEmpty;

  @override
  bool get isNotEmpty => _data.isNotEmpty;

  @override
  Iterator<E> get iterator => _data.iterator;

  @override
  String join([String separator = '']) => _data.join(separator);

  @override
  E get last => _data.last;

  @override
  E lastWhere(bool Function(E) test, {E Function()? orElse}) => _data.lastWhere(test, orElse: orElse);

  @override
  int get length => _data.length;

  @override
  Iterable<T> map<T>(T Function(E) f) => _data.map(f);

  @override
  E reduce(E Function(E value, E element) combine) => _data.reduce(combine);

  @override
  E get single => _data.single;

  @override
  E singleWhere(bool Function(E) test, {E Function()? orElse}) => _data.singleWhere(test, orElse: orElse);

  @override
  Iterable<E> skip(int n) => _data.skip(n);

  @override
  Iterable<E> skipWhile(bool Function(E) test) => _data.skipWhile(test);

  @override
  Iterable<E> take(int n) => _data.take(n);

  @override
  Iterable<E> takeWhile(bool Function(E) test) => _data.takeWhile(test);

  @override
  List<E> toList({bool growable = true}) => _data.toList(growable: growable);

  @override
  Set<E> toSet() => _data.toSet();

  @override
  Iterable<E> where(bool Function(E) test) => _data.where(test);

  @override
  Iterable<T> whereType<T>() => _data.whereType<T>();

  @override
  String toString() => _data.toString();

  @override
  bool add(E value) {
    final result = _data.add(value);
    if (result) invalidate();
    return result;
  }

  @override
  void addAll(Iterable<E> iterable) {
    if (iterable.isEmpty) return;
    final length = _data.length;
    _data.addAll(iterable);
    if (length != _data.length) invalidate();
  }

  @override
  void clear() {
    if (_data.isEmpty) return;
    _data.clear();
    invalidate();
  }

  @override
  bool containsAll(Iterable<Object?> other) => _data.containsAll(other);

  @override
  Set<E> difference(Set<Object?> other) => _data.difference(other);

  @override
  Set<E> intersection(Set<Object?> other) => _data.intersection(other);

  @override
  E? lookup(Object? element) => _data.lookup(element);

  @override
  bool remove(Object? value) {
    final result = _data.remove(value);
    if (result) invalidate();
    return result;
  }

  @override
  void removeAll(Iterable<Object?> elements) {
    final length = _data.length;
    _data.removeAll(elements);
    if (length != _data.length) invalidate();
  }

  @override
  void removeWhere(bool Function(E) test) {
    final length = _data.length;
    _data.removeWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  void retainAll(Iterable<Object?> elements) {
    final length = _data.length;
    _data.retainAll(elements);
    if (length != _data.length) invalidate();
  }

  @override
  void retainWhere(bool Function(E) test) {
    final length = _data.length;
    _data.retainWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  Set<E> union(Set<E> other) => _data.union(other);
}

/// [MediatorMixin] implementation of [LiveSet]
class MediatorLiveSet<E> extends LiveSet<E> with MediatorMixin<Set<E>> {
  @override
  final bool reactive;

  /// Check [LiveSet.new]
  MediatorLiveSet({Set<E>? source, this.reactive = true}) : super(source: source);
}

/// [HistoryMixin] implementation of [LiveSet]
class HistoryLiveSet<E> extends LiveSet<E> with HistoryMixin<Set<E>> {
  /// Check [LiveSet.new]
  HistoryLiveSet({Set<E>? source}) : super(source: source);
}

/// [StorageMixin] implementation of [LiveSet]
class StorageLiveSet<E> extends LiveSet<E> with StorageMixin<Set<E>> {
  @override
  final String key;

  @override
  final Set<E> defaultValue;

  @override
  final Map<String, String> storage;

  @override
  final String Function(Set<E> value) encode;

  /// Check [StorageMixin]
  StorageLiveSet({
    required this.key,
    required this.storage,
    Set<E>? defaultValue,
    this.encode = StorageMixin.encodeJsonIterable,
    Set<E> Function(String source)? decode,
  })  : defaultValue = defaultValue ?? const {},
        super(source: StorageMixin.resolve(key, storage, defaultValue?.toSet(), decode ?? _jsonDecodeSet));

  /// Check [StorageMixin]
  StorageLiveSet.enumeration({
    required String key,
    required Map<String, String> storage,
    Set<E>? defaultValue,
    required Iterable<E> values,
  }) : this(
            key: key,
            storage: storage,
            defaultValue: defaultValue,
            encode: StorageMixin.encodeEnumIterable,
            decode: (source) => StorageMixin.decodeEnumIterable<E>(source, values).toSet());

  /// Check [StorageMixin]
  StorageLiveSet.jsonObject({
    required String key,
    required Map<String, String> storage,
    Set<E>? defaultValue,
    dynamic Function(E value)? toJson,
    required E Function(Map<String, dynamic> json) fromJson,
  }) : this(
            key: key,
            storage: storage,
            defaultValue: defaultValue,
            encode: (value) => StorageMixin.encodeJsonIterable(toJson == null ? value : value.map(toJson)),
            decode: (source) => StorageMixin.decodeJsonIterable<E>(source, (item) => fromJson(item)).toSet());

  static Set<E> _jsonDecodeSet<E>(String source) => Set<E>.from(jsonDecode(source));
}

/// A [LiveCollection] which implements [Queue]
class LiveQueue<E> extends LiveCollection<Queue<E>> implements Queue<E> {
  /// Create from [source]
  LiveQueue({Queue<E>? source}) : super(source ?? Queue<E>());

  @override
  bool shouldUpdateValue(Queue<E> oldValue, Queue<E> newValue) {
    if (identical(oldValue, newValue)) return false;
    if (oldValue.length != newValue.length) return true;
    final a = oldValue.iterator;
    final b = newValue.iterator;
    while (a.moveNext() && b.moveNext()) if (a.current != b.current) return true;
    return false;
  }

  @override
  void setTo(Iterable<E> value) {
    _data
      ..clear()
      ..addAll(value);
    invalidate();
  }

  @override
  bool any(bool Function(E) test) => _data.any(test);

  @override
  Queue<T> cast<T>() => _data.cast<T>();

  @override
  bool contains(Object? element) => _data.contains(element);

  @override
  E elementAt(int index) => _data.elementAt(index);

  @override
  bool every(bool Function(E) test) => _data.every(test);

  @override
  Iterable<T> expand<T>(Iterable<T> Function(E) f) => _data.expand(f);

  @override
  E get first => _data.first;

  @override
  E firstWhere(bool Function(E) test, {E Function()? orElse}) => _data.firstWhere(test, orElse: orElse);

  @override
  T fold<T>(T initialValue, T Function(T previousValue, E element) combine) => _data.fold(initialValue, combine);

  @override
  Iterable<E> followedBy(Iterable<E> other) => _data.followedBy(other);

  @override
  void forEach(void Function(E) f) => _data.forEach(f);

  @override
  bool get isEmpty => _data.isEmpty;

  @override
  bool get isNotEmpty => _data.isNotEmpty;

  @override
  Iterator<E> get iterator => _data.iterator;

  @override
  String join([String separator = '']) => _data.join(separator);

  @override
  E get last => _data.last;

  @override
  E lastWhere(bool Function(E) test, {E Function()? orElse}) => _data.lastWhere(test, orElse: orElse);

  @override
  int get length => _data.length;

  @override
  Iterable<T> map<T>(T Function(E) f) => _data.map(f);

  @override
  E reduce(E Function(E value, E element) combine) => _data.reduce(combine);

  @override
  E get single => _data.single;

  @override
  E singleWhere(bool Function(E) test, {E Function()? orElse}) => _data.singleWhere(test, orElse: orElse);

  @override
  Iterable<E> skip(int n) => _data.skip(n);

  @override
  Iterable<E> skipWhile(bool Function(E) test) => _data.skipWhile(test);

  @override
  Iterable<E> take(int n) => _data.take(n);

  @override
  Iterable<E> takeWhile(bool Function(E) test) => _data.takeWhile(test);

  @override
  List<E> toList({bool growable = true}) => _data.toList(growable: growable);

  @override
  Set<E> toSet() => _data.toSet();

  @override
  Iterable<E> where(bool Function(E) test) => _data.where(test);

  @override
  Iterable<T> whereType<T>() => _data.whereType<T>();

  @override
  String toString() => _data.toString();

  @override
  void add(E value) {
    _data.add(value);
    invalidate();
  }

  @override
  void addAll(Iterable<E> iterable) {
    if (iterable.isEmpty) return;
    _data.addAll(iterable);
    invalidate();
  }

  @override
  void addFirst(E value) {
    _data.addFirst(value);
    invalidate();
  }

  @override
  void addLast(E value) {
    _data.addLast(value);
    invalidate();
  }

  @override
  void clear() {
    if (_data.isEmpty) return;
    _data.clear();
    invalidate();
  }

  @override
  bool remove(Object? value) {
    final result = _data.remove(value);
    if (result) invalidate();
    return result;
  }

  @override
  E removeFirst() {
    final result = _data.removeFirst();
    invalidate();
    return result;
  }

  @override
  E removeLast() {
    final result = _data.removeLast();
    invalidate();
    return result;
  }

  @override
  void removeWhere(bool Function(E) test) {
    final length = _data.length;
    _data.removeWhere(test);
    if (length != _data.length) invalidate();
  }

  @override
  void retainWhere(bool Function(E) test) {
    final length = _data.length;
    _data.retainWhere(test);
    if (length != _data.length) invalidate();
  }
}

/// [MediatorMixin] implementation of [LiveQueue]
class MediatorLiveQueue<E> extends LiveQueue<E> with MediatorMixin<Queue<E>> {
  @override
  final bool reactive;

  /// Check [LiveQueue.new]
  MediatorLiveQueue({Queue<E>? source, this.reactive = true}) : super(source: source);
}

/// [HistoryMixin] implementation of [LiveQueue]
class HistoryLiveQueue<E> extends LiveQueue<E> with HistoryMixin<Queue<E>> {
  /// Check [LiveQueue.new]
  HistoryLiveQueue({Queue<E>? source}) : super(source: source);
}

/// [StorageMixin] implementation of [LiveQueue]
class StorageLiveQueue<E> extends LiveQueue<E> with StorageMixin<Queue<E>> {
  @override
  final String key;

  @override
  final Queue<E> defaultValue;

  @override
  final Map<String, String> storage;

  @override
  final String Function(Queue<E> value) encode;

  /// Check [StorageMixin]
  StorageLiveQueue({
    required this.key,
    required this.storage,
    Queue<E>? defaultValue,
    this.encode = StorageMixin.encodeJsonIterable,
    Queue<E> Function(String source)? decode,
  })  : defaultValue = defaultValue ?? Queue(),
        super(
            source: StorageMixin.resolve(
                key, storage, defaultValue == null ? null : Queue.of(defaultValue), decode ?? _jsonDecodeQueue));

  /// Check [StorageMixin]
  StorageLiveQueue.enumeration({
    required this.key,
    required this.storage,
    Queue<E>? defaultValue,
    required Iterable<E> values,
  })  : encode = StorageMixin.encodeEnumIterable,
        defaultValue = defaultValue ?? Queue(),
        super(
            source: StorageMixin.resolve(key, storage, defaultValue == null ? null : Queue.of(defaultValue),
                (source) => Queue.of(StorageMixin.decodeEnumIterable<E>(source, values))));

  /// Check [StorageMixin]
  StorageLiveQueue.jsonObject({
    required String key,
    required Map<String, String> storage,
    Queue<E>? defaultValue,
    dynamic Function(E value)? toJson,
    required E Function(Map<String, dynamic> json) fromJson,
  }) : this(
            key: key,
            storage: storage,
            defaultValue: defaultValue,
            encode: (value) => StorageMixin.encodeJsonIterable(toJson == null ? value : value.map(toJson)),
            decode: (source) => Queue.of(StorageMixin.decodeJsonIterable<E>(source, (item) => fromJson(item))));

  static Queue<E> _jsonDecodeQueue<E>(String source) => Queue<E>.from(jsonDecode(source));
}
