part of '../flutterx_live_data.dart';

/// Some useful extensions
extension LiveDataExt<X> on LiveData<X> {
  /// Map this LiveData to a new LiveData which depends on this one
  LiveData<Y> liveMap<Y>(FutureOr<Y> Function(X) transform, {bool reactive = true}) =>
      MediatorLiveData<Y>(reactive: reactive)..addTransformSource<X>(this, transform);

  /// Observe within [StateObserver] context
  void observe(StateObserver context, Observer<X> onChanged) => context._observe<X>(this, onChanged);

  /// Converts this LiveData into a [Future] which completes when it's value matches [accept] predicate.
  /// You can also set a [timeLimit] after which if the future has not yet completed, will complete with
  /// [onTimeout] result if specified or throw a [TimeoutException] if omitted
  Future<X> whenValue({bool Function(X)? accept, Duration? timeLimit, FutureOr<X> Function()? onTimeout}) {
    final completer = Completer<X>();
    void _observe(X value) => accept == null || accept(value) ? completer.complete(value) : null;
    observeForever(_observe);
    return (timeLimit == null ? completer.future : completer.future.timeout(timeLimit, onTimeout: onTimeout))
      ..whenComplete(() => removeObserver(_observe));
  }
}

/// Shortcuts for transforming multiple LiveData sources into the mediator target
extension MediatorExt<Y> on MediatorMixin<Y> {
  /// Check [addTransformSources]
  void addTransformSource2<X1, X2>(LiveData<X1> x1, LiveData<X2> x2, FutureOr<Y> Function(X1, X2) transform) =>
      addTransformSources([x1, x2], (values) => transform(values[0], values[1]));

  /// Check [addTransformSources]
  void addTransformSource3<X1, X2, X3>(
          LiveData<X1> x1, LiveData<X2> x2, LiveData<X3> x3, FutureOr<Y> Function(X1, X2, X3) transform) =>
      addTransformSources([x1, x2, x3], (values) => transform(values[0], values[1], values[2]));

  /// Check [addTransformSources]
  void addTransformSource4<X1, X2, X3, X4>(LiveData<X1> x1, LiveData<X2> x2, LiveData<X3> x3, LiveData<X4> x4,
          FutureOr<Y> Function(X1, X2, X3, X4) transform) =>
      addTransformSources([x1, x2, x3, x4], (values) => transform(values[0], values[1], values[2], values[4]));
}

/// Shortcuts for observing multiple LiveData simultaneously
extension StateObserverExt on StateObserver {
  /// Check [observeAll]
  void observe2<X1, X2>(LiveData<X1> x1, LiveData<X2> x2, void Function(X1, X2) onChanged) =>
      observeAll([x1, x2], (values) => onChanged(values[0], values[1]));

  /// Check [observeAll]
  void observe3<X1, X2, X3>(LiveData<X1> x1, LiveData<X2> x2, LiveData<X3> x3, void Function(X1, X2, X3) onChanged) =>
      observeAll([x1, x2, x3], (values) => onChanged(values[0], values[1], values[2]));

  /// Check [observeAll]
  void observe4<X1, X2, X3, X4>(LiveData<X1> x1, LiveData<X2> x2, LiveData<X3> x3, LiveData<X4> x4,
          void Function(X1, X2, X3, X4) onChanged) =>
      observeAll([x1, x2, x3, x4], (values) => onChanged(values[0], values[1], values[2], values[3]));
}

/// Check [observeAllForever]
void observe2Forever<X1, X2>(LiveData<X1> x1, LiveData<X2> x2, void Function(X1, X2) onChanged) =>
    observeAllForever([x1, x2], (values) => onChanged(values[0], values[1]));

/// Check [observeAllForever]
void observe3Forever<X1, X2, X3>(
        LiveData<X1> x1, LiveData<X2> x2, LiveData<X3> x3, void Function(X1, X2, X3) onChanged) =>
    observeAllForever([x1, x2, x3], (values) => onChanged(values[0], values[1], values[2]));

/// Check [observeAllForever]
void observe4Forever<X1, X2, X3, X4>(
        LiveData<X1> x1, LiveData<X2> x2, LiveData<X3> x3, LiveData<X4> x4, void Function(X1, X2, X3, X4) onChanged) =>
    observeAllForever([x1, x2, x3, x4], (values) => onChanged(values[0], values[1], values[2], values[3]));

/// Observe multiple sources simultaneously. Check [StateObserverExt]
List<ObserverWrapper> observeAllForever(Iterable<LiveData> sources, void Function(List values) onChanged) {
  final result = <ObserverWrapper>[];
  _observeAllInternal(sources, onChanged, <V>(source, onChanged) => result.add(source.observeForever(onChanged)));
  return result;
}

void _observeAllInternal(Iterable<LiveData> sources, void Function(List values) onChanged,
    void Function<V>(LiveData<V> source, Observer<V> onChanged) observe) {
  sources = sources.toList(growable: false);
  var ready = false;
  var notify = false;
  for (final data in sources)
    observe(data, (_) {
      List values;
      try {
        values = sources.map((data) => data._data).toList(growable: false);
      } catch (_) {
        return; // at least one source is not initialized
      }
      if (ready) {
        onChanged(values);
      } else {
        notify = true;
      }
    });
  if (notify) onChanged(sources.map((data) => data._data).toList(growable: false));
  ready = true;
}
