part of '../flutterx_live_data.dart';

/// LiveData is a data holder class that can be observed.
/// This is a Flutter implementation of LiveData in Android Jetpack library
///
/// A LiveData has a value which by default is uninitialized.
/// If the current value of this LiveData has been initialized, it will be dispatched to every observer
/// that becomes active
///
/// An [Observer] can be added and will be notified about modifications of the wrapped data only
/// if is in active state.
/// An observer added via [observeForever] is considered as always active and thus will be always notified
/// about modifications. For those observers, you should manually call [removeObserver]
///
/// To observe a LiveData inside a [State] you should implement [StateObserver] mixin on that state,
/// override [StateObserver.registerObservers] and here you can call [LiveDataExt.observe]
///
/// Alternatively You can use [LiveDataBuilder] and [CombinedLiveDataBuilder] to handle live data changes
/// withing the scope of a widget tree. The widget is automatically rebuilt every time the data is updated
///
/// Each modification can happen through protected [setData] and [postData] methods
/// The first one sets the data immediately and [invalidate] this live data,
/// while [postData] calls [setData] after future task completion
///
/// [setData] will update the current value and invalidates this LiveData only if [shouldUpdateValue] returns true
/// This check changes from one implementation to another, but by default is newValue != oldValue
/// This means that if you set the same value twice consecutively only the first time it will be dispatched to the
/// observers
///
/// In addition, LiveData has [onActive] and [onInactive] methods
/// to get notified when number of active [Observer]s change between 0 and 1.
/// This allows LiveData to release any heavy resources when it does not have any Observers that
/// are actively observing.
abstract class LiveData<T> {
  final Map<Observer<T>, ObserverWrapper<T>> _observers = {};
  late T _data;
  int _version = -1;
  int _activeCount = 0;

  /// Current [value] version: -1 if uninitialized, this value is incremented every time [invalidate] is called
  int get version => _version;

  /// True if [value] has been initialized. If false any reference to [value] will throw an exception
  bool get initialized => _version > -1;

  /// True if this LiveData currently has active observers
  bool get isActive => _activeCount > 0;

  /// Actual value of this LiveData
  T get value => _data;

  /// Called when [isActive] changes to true
  @protected
  void onActive() {}

  /// Called when [isActive] changes to false
  @protected
  void onInactive() {}

  /// Check if [oldValue] should be updated with [newValue]
  @protected
  bool shouldUpdateValue(T oldValue, T newValue) => newValue != oldValue;

  /// Calls [setData] after future task completion
  @protected
  void postData(FutureOr<T> value) => value is Future<T> ? value.then(setData) : setData(value);

  /// Update current [LiveData.value] with new [value] and [invalidate]s only if [shouldUpdateValue] return true
  @protected
  void setData(T value) {
    if (!initialized || shouldUpdateValue(_data, value)) {
      _data = value;
      invalidate();
    }
  }

  /// Increment [version] and notifies all active observers that the data is changed
  void invalidate() {
    _version++;
    _dispatchingValue(null);
  }

  /// Observe this LiveData until observer is manually removed. This method will call [addObserverWrapper] passing
  /// an [ObserverWrapper] instance created with [onChanged] and [dispatcher] parameters.
  ObserverWrapper<T> observeForever(Observer<T> onChanged, {Dispatcher? dispatcher}) =>
      addObserverWrapper(ObserverWrapper<T>(this, onChanged, dispatcher));

  /// Removes the [observer] from this LiveData
  void removeObserver(Observer<T> observer) => _observers.remove(observer)?.onRemove();

  /// Adds an [observer] to this LiveData
  ObserverWrapper<T> addObserverWrapper(ObserverWrapper<T> observer) {
    assert(!_observers.containsKey(observer.onChanged), 'observer already registered');
    _observers[observer.onChanged] = observer;
    return observer..onAdded();
  }

  /// Removes the [observer] from this LiveData from [ObserverWrapper] instance
  void removeObserverWrapper(ObserverWrapper<T> observer) => removeObserver(observer.onChanged);

  void _changeActiveState(ObserverWrapper<T> observer, {required bool active}) {
    if (active) {
      if (++_activeCount == 1) onActive();
      _dispatchingValue(observer);
    } else if (--_activeCount == 0) onInactive();
  }

  void _dispatchingValue(ObserverWrapper<T>? initiator) {
    if (initialized) {
      if (initiator != null)
        initiator._considerNotify();
      else
        for (final observer in _observers.values.toList(growable: false)) observer._considerNotify();
    }
  }

  @override
  String toString() =>
      '${objectRuntimeType(this, 'LiveData')}${initialized ? (identical(value, this) ? '' : '($value)') : '[uninitialized]'})';
}

/// [StorageMixin] implementation of [LiveData]
class MediatorLiveData<T> extends LiveData<T> with MediatorMixin<T> {
  @override
  final bool reactive;

  /// Check [LiveData.new]
  MediatorLiveData({this.reactive = true});
}
