part of '../flutterx_live_data.dart';

/// A [SelfLiveData] instance which represents a stateful result. States are [ResultState.values]
class LiveResult<T> extends SelfLiveData<LiveResult<T>> {
  final MutableLiveData<ResultState> _state = MutableLiveData(initialValue: ResultState.idle);
  dynamic _value;

  /// Current state value
  ResultState get state => _state._data;

  /// Result value only available if [state] is [ResultState.success]
  T get result => _state._data == ResultState.success
      ? _value
      : throw StateError('result can only be accessed when state is ResultState.success');

  /// Error value only available if [state] is [ResultState.error]
  Object get error => _state._data == ResultState.error
      ? _value
      : throw StateError('error can only be accessed when state is ResultState.error');

  /// Default constructor
  LiveResult() {
    addField(_state);
  }

  /// Automatically sync events with given [task]
  /// Same as calling setLoading(); try { setResult(await task); } catch (e) { setError(e); }
  void load(Future<T> task) {
    setLoading();
    task.then(setResult, onError: setError);
  }

  /// Set state to [ResultState.loading]
  void setLoading() {
    _value = null;
    _state.setData(ResultState.loading);
  }

  /// Set state to [ResultState.success] and set [result] to [value]
  void setResult(T value) {
    _value = value;
    _state.setData(ResultState.success);
  }

  /// Set state to [ResultState.error] and set [LiveResult.error] to [error]
  void setError(dynamic error) {
    _value = error;
    _state.setData(ResultState.error);
  }

  /// Reset state to [ResultState.idle]
  void reset() {
    _value = null;
    _state.setData(ResultState.idle);
  }
}

/// States of [LiveResult]
enum ResultState {
  /// Task has not yet been requested
  idle,

  /// Task has started and result is loading
  loading,

  /// Task has completed with an error
  error,

  /// Task has completed successfully
  success,
}
