part of '../flutterx_live_data.dart';

/// A [LiveData] mixin which may observe other [LiveData] objects and react on [Observer] events from them.
/// This class correctly propagates its active/inactive states down to source [LiveData] objects.
mixin MediatorMixin<Y> on LiveData<Y> {
  final Map<LiveData, _Source> _sources = {};

  /// If false, this LiveData will observe sources only if it's currently active
  bool get reactive;

  @override
  void onActive() {
    super.onActive();
    if (!reactive) _plugAll();
  }

  @override
  void onInactive() {
    if (!reactive) _unplugAll();
    super.onInactive();
  }

  /// Add mediator sources which apply a combined transform mapping to it's source values. Check [MediatorExt]
  void addTransformSources(Iterable<LiveData> sources, FutureOr<Y> Function(List values) transform) =>
      _observeAllInternal(sources, (values) => postData(transform(values)), addSource);

  /// Add mediator source which apply a simple transform mapping to it's source value
  void addTransformSource<X>(LiveData<X> source, FutureOr<Y> Function(X) transform) =>
      addSource<X>(source, (value) => postData(transform(value)));

  /// Add mediator source which uses the transformed live data as mediator source
  void addLiveTransformSource<X>(LiveData<X> source, LiveData<Y> Function(X) transform) {
    LiveData<Y>? lastSource;
    addSource<X>(source, (value) {
      final newSource = transform(value);
      if (identical(lastSource, newSource)) return;
      if (lastSource != null) removeSource(lastSource!);
      addSource(lastSource = newSource, setData);
    });
  }

  /// Add the given [source] to the mediator sources list
  void addSource<X>(LiveData<X> source, Observer<X> onChanged) {
    assert(!_sources.containsKey(source), 'source already registered');
    final wrapper = _Source<X>(source, onChanged);
    _sources[source] = wrapper;
    if (isActive || reactive) wrapper.plug();
  }

  /// Removes the given [source] from the mediator sources list.
  void removeSource(LiveData source) => _sources.remove(source)?.unplug();

  /// Removes the given [sources] from the mediator sources list.
  void removeSources(Iterable<LiveData> sources) => sources.toList(growable: false).forEach(removeSource);

  void _plugAll() {
    for (final source in _sources.values) source.plug();
  }

  void _unplugAll() {
    for (final source in _sources.values) source.unplug();
  }
}

class _Source<X> {
  final LiveData<X> data;

  final Observer<X> onChanged;

  int lastVersion = -1;

  _Source(this.data, this.onChanged);

  void plug() => data.observeForever(_onChanged);

  void unplug() => data.removeObserver(_onChanged);

  void _onChanged(X value) {
    if (lastVersion < data._version) {
      lastVersion = data._version;
      onChanged(value);
    }
  }
}
