part of '../flutterx_live_data.dart';

/// Type alias for observing data changes
typedef Observer<T> = void Function(T value);

/// Wrapper of [Observer] function
class ObserverWrapper<T> {
  /// LiveData which is observed
  @protected
  final LiveData<T> data;

  /// Callback observer function
  @protected
  final Observer<T> onChanged;

  /// Optional dispatcher for [onChanged] callback
  @protected
  final Dispatcher? dispatcher;

  bool _active = false;

  /// True if the observer is currently registered to [data] changes
  bool get active => _active;

  int _lastVersion = -1;

  /// Create an Observer of [data]
  ObserverWrapper(this.data, this.onChanged, this.dispatcher);

  /// Called when observer is added to [data]
  @protected
  @mustCallSuper
  void onAdded() => _changeActiveState(active: true);

  /// Called when observer is removed from [data]
  @protected
  @mustCallSuper
  void onRemove() => _changeActiveState(active: false);

  /// Remove this observer from currently active [data]
  void dispose() => data.removeObserver(onChanged);

  void _changeActiveState({required bool active}) {
    if (_active == active) return;
    _active = active;
    data._changeActiveState(this, active: active);
  }

  late T _pendingData;
  bool _dispatchingPendingData = false;

  void _considerNotify() {
    if (_active && data.version > _lastVersion) {
      _lastVersion = data.version;
      final dispatcher = this.dispatcher;
      if (dispatcher == null) return onChanged(data._data);
      _pendingData = data._data;
      if (_dispatchingPendingData) return;
      _dispatchingPendingData = true;
      dispatcher.post(() {
        if (_active) onChanged(_pendingData);
        _dispatchingPendingData = false;
      });
    }
  }
}
