part of '../flutterx_live_data.dart';

/// Link [LiveData] observers to the lifecycle of your [State]
///
/// Implement this mixin on your widget's [State] to add [Observer]s to a [LiveData] within the state lifecycle
/// Every LiveData can be observed inside [registerObservers] method that you can override
/// The observer must be added through proper extension methods and will be removed automatically on state [dispose]
/// Check [StateObserverExt]
mixin StateObserver<T extends StatefulWidget> on State<T> {
  final Map<_ObserverKey, ObserverWrapper> _disposables = {};
  Set<_ObserverKey>? _keys;

  @override
  void initState() {
    super.initState();
    doRegister();
  }

  @override
  void didUpdateWidget(T oldWidget) {
    super.didUpdateWidget(oldWidget);
    doRegister();
  }

  @override
  void dispose() {
    doUnregister();
    super.dispose();
  }

  /// Override this method to observe you LiveData(s)
  @protected
  void registerObservers();

  /// Trigger calling [registerObservers] and automatically manage registering/disposing observers
  @protected
  void doRegister() {
    final keys = _keys = _disposables.keys.toSet();
    registerObservers();
    for (final key in keys) _disposables.remove(key)!.dispose();
    _keys = null;
  }

  /// Dispose all active observers
  @protected
  void doUnregister() {
    for (final key in _disposables.keys.toList(growable: false)) _disposables.remove(key)!.dispose();
  }

  /// Observe multiple sources simultaneously within this context. Check [StateObserverExt]
  @protected
  void observeAll(Iterable<LiveData> sources, void Function(List values) onChanged) =>
      _observeAllInternal(sources, onChanged, _observe);

  void _observe<V>(LiveData<V> data, Observer<V> onChanged) {
    final keys = _keys;
    if (keys == null) throw StateError('Observe can only be called within registerObservers function');
    final key = _ObserverKey(data, onChanged);
    _disposables.containsKey(key) ? keys.remove(key) : _disposables[key] = data.observeForever(onChanged);
  }
}

@immutable
class _ObserverKey {
  final Object _data;
  final Object _onChanged;

  const _ObserverKey(this._data, this._onChanged);

  @override
  bool operator ==(Object other) =>
      other is _ObserverKey && identical(other._data, _data) && other._onChanged == _onChanged;

  @override
  int get hashCode => hashValues(_data, _onChanged);
}
