part of '../flutterx_live_data.dart';

/// Builds a [Widget] depending on [LiveData] observed value
typedef LiveWidgetBuilder<T> = Widget Function(BuildContext context, T value);

/// A [Widget] whose content is built from the observed [LiveData]
class LiveDataBuilder<T> extends StatelessWidget {
  /// Data which is observed
  final LiveData<T> data;

  /// Widget builder function
  final LiveWidgetBuilder<T> builder;

  /// Displayed when [data] is uninitialized, defaults to [SizedBox.shrink]
  final Widget? placeholder;

  /// Creates [builder] from [data]
  const LiveDataBuilder({Key? key, required this.data, required this.builder, this.placeholder}) : super(key: key);

  @override
  StatelessElement createElement() => _LiveDataBuilderElement<T>(this);

  @override
  Widget build(BuildContext context) => _liveBuild<T>(context, data, builder, placeholder);

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) => super.debugFillProperties(properties
    ..add(DiagnosticsProperty<LiveData<T>>('data', data))
    ..add(ObjectFlagProperty<LiveWidgetBuilder<T>>.has('builder', builder)));
}

class _LiveDataBuilderElement<T> extends StatelessElement with _LiveDataBuilderMixin<T, LiveDataBuilder<T>> {
  _LiveDataBuilderElement(LiveDataBuilder<T> widget) : super(widget);

  @override
  bool dataChanged(LiveDataBuilder<T> oldWidget, LiveDataBuilder<T> newWidget) =>
      !identical(oldWidget.data, newWidget.data);

  @override
  void link(LiveDataBuilder<T> widget) => widget.data.observeForever(onChanged, dispatcher: Dispatcher.postFrame);

  @override
  void unlink(LiveDataBuilder<T> widget) => widget.data.removeObserver(onChanged);
}

Widget _liveBuild<T>(BuildContext context, LiveData<T> data, LiveWidgetBuilder<T> builder, Widget? placeholder) =>
    data.initialized ? builder(context, data._data) : (placeholder ?? const SizedBox.shrink());

mixin _LiveDataBuilderMixin<T, W extends StatelessWidget> on StatelessElement {
  @override
  W get widget => super.widget as W;

  @override
  void mount(Element? parent, dynamic newSlot) {
    link(widget);
    super.mount(parent, newSlot);
  }

  @override
  void update(W newWidget) {
    final oldWidget = widget;
    if (dataChanged(oldWidget, newWidget)) {
      unlink(oldWidget);
      link(newWidget);
    }
    super.update(newWidget);
  }

  @override
  void unmount() {
    unlink(widget);
    super.unmount();
  }

  @protected
  bool dataChanged(W oldWidget, W newWidget);

  @protected
  void link(W widget);

  @protected
  void unlink(W widget);

  @protected
  void onChanged(_) => markNeedsBuild();
}
