part of '../flutterx_live_data.dart';

/// A [Widget] whose content is built from the observed [LiveData] list
class CombinedLiveDataBuilder<T> extends StatelessWidget {
  /// Data list which is observed
  final List<LiveData> data;

  /// Computes value from [data] sources
  final T Function(List values) transform;

  /// Widget builder function
  final LiveWidgetBuilder<T> builder;

  /// Displayed when [data] is uninitialized, defaults to [SizedBox.shrink]
  final Widget? placeholder;

  /// Creates [builder] from [transform]ed [data]
  const CombinedLiveDataBuilder({
    Key? key,
    required this.data,
    required this.transform,
    required this.builder,
    this.placeholder,
  }) : super(key: key);

  /// Check [CombinedLiveDataBuilder.new]
  static CombinedLiveDataBuilder<Y> with2<X1, X2, Y>(
          {Key? key,
          required LiveData<X1> x1,
          required LiveData<X2> x2,
          required Y Function(X1, X2) transform,
          required LiveWidgetBuilder<Y> builder,
          Widget? placeholder}) =>
      CombinedLiveDataBuilder<Y>(
          data: [x1, x2],
          transform: (values) => transform(values[0], values[1]),
          builder: builder,
          placeholder: placeholder);

  /// Check [CombinedLiveDataBuilder.new]
  static CombinedLiveDataBuilder<Y> with3<X1, X2, X3, Y>(
          {Key? key,
          required LiveData<X1> x1,
          required LiveData<X2> x2,
          required LiveData<X3> x3,
          required Y Function(X1, X2, X3) transform,
          required LiveWidgetBuilder<Y> builder,
          Widget? placeholder}) =>
      CombinedLiveDataBuilder<Y>(
          data: [x1, x2, x3],
          transform: (values) => transform(values[0], values[1], values[2]),
          builder: builder,
          placeholder: placeholder);

  /// Check [CombinedLiveDataBuilder.new]
  static CombinedLiveDataBuilder<Y> with4<X1, X2, X3, X4, Y>(
          {Key? key,
          required LiveData<X1> x1,
          required LiveData<X2> x2,
          required LiveData<X3> x3,
          required LiveData<X4> x4,
          required Y Function(X1, X2, X3, X4) transform,
          required LiveWidgetBuilder<Y> builder,
          Widget? placeholder}) =>
      CombinedLiveDataBuilder<Y>(
          data: [x1, x2, x3, x4],
          transform: (values) => transform(values[0], values[1], values[2], values[3]),
          builder: builder,
          placeholder: placeholder);

  @override
  StatelessElement createElement() => _CombinedLiveDataBuilderElement<T>(this);

  @override
  Widget build(BuildContext context) =>
      _liveBuild<T>(context, (context as _CombinedLiveDataBuilderElement<T>)._mediator, builder, placeholder);

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) => super.debugFillProperties(properties
    ..add(IterableProperty<LiveData>('data', data))
    ..add(ObjectFlagProperty<LiveWidgetBuilder<T>>.has('builder', builder))
    ..add(ObjectFlagProperty<T Function(List values)>.has('transform', transform)));
}

class _CombinedLiveDataBuilderElement<T> extends StatelessElement
    with _LiveDataBuilderMixin<T, CombinedLiveDataBuilder<T>> {
  final MediatorLiveData<T> _mediator = MediatorLiveData<T>(reactive: false);

  _CombinedLiveDataBuilderElement(CombinedLiveDataBuilder<T> widget) : super(widget);

  @override
  bool dataChanged(CombinedLiveDataBuilder<T> oldWidget, CombinedLiveDataBuilder<T> newWidget) =>
      !listEquals(oldWidget.data, newWidget.data);

  @override
  void link(CombinedLiveDataBuilder<T> widget) => _mediator
    ..addTransformSources(widget.data, widget.transform)
    ..observeForever(onChanged, dispatcher: Dispatcher.postFrame);

  @override
  void unlink(CombinedLiveDataBuilder<T> widget) => _mediator
    ..removeObserver(onChanged)
    ..removeSources(widget.data);
}
